
/*****************************************************************************************************************************
Brushless Gimbal Controller Software by Christian Winkler and Alois Hahn (C) 2013

Brushless Gimbal Controller Hardware and Software support 
by Ludwig Fäerber, Alexander Rehfeld and Martin Eckart

Special Contributions:
  Michael Schätzel

Project homepage: http://brushlessgimbal.de/
Discussions:
http://fpv-community.de/showthread.php?20795-Brushless-Gimbal-Controller-SOFTWARE
http://fpv-community.de/showthread.php?22617-Gimbal-Brushless-Controller-V3-0-50x50mm-by-Martinez
http://fpv-community.de/showthread.php?19252-Brushless-Gimbal-Controller

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version. see <http://www.gnu.org/licenses/>

Anyhow, if you start to commercialize our work, please read on http://code.google.com/p/brushless-gimbal/ on how to contribute

// I2Cdev library collection - MPU6050 I2C device class
// Based on InvenSense MPU-6050 register map document rev. 2.0, 5/19/2011 (RM-MPU-6000A-00)
// 10/3/2011 by Jeff Rowberg <jeff@rowberg.net>
// Updates should (hopefully) always be available at https://github.com/jrowberg/i2cdevlib
*****************************************************************************************************************************/

/* Note: software has been tested with Arduino Version 1.5.4 */

#define VERSION_STATUS B // A = Alpha; B = Beta , N = Normal Release
#define VERSION "v50"
#define REVISION "r207"
#define VERSION_EEPROM 15 // change this number when eeprom data structure has changed


/*************************/
/* Include Header Files  */
/*************************/
#include <EEPROM.h>
#include <Wire.h>
#include <avr/pgmspace.h>
#include "definitions.h"
#include "MPU6050.h"
#include "SerialCommand.h"
#include "EEPROMAnything.h"
#include "PinChangeInt.h"
#include "Timer1.h"
#include "Trace.h"
#include "variables.h"
MPU6050 mpu;            // Create MPU object
SerialCommand sCmd;     // Create SerialCommand object

#include "fastMathRoutines.h"     // fast Math functions required by orientationRoutines.h
#include "orientationRoutines.h"  // get Orientation from ACC
#include "RCdecode.h"             // RC Decoder to move camera by servo signal input
#include "BLcontroller.h"         // Motor Movement Functions and Timer Config
#include "SerialCom.h"            // Serial Protocol for Configuration and Communication


/**********************************************/
/* Initialization                             */
/**********************************************/
void setup() 
{

  // just for debugging
#ifdef STACKHEAPCHECK_ENABLE
  stackCheck();
  heapCheck();
#endif

  LEDPIN_PINMODE
  LEDPIN2_PINMODE
  
  CH2_PINMODE
  CH3_PINMODE

  // Start Serial Port
  Serial.begin(115200);

  // send Version Number and welcome message
  printMessage(MSG_INFO, F("BruGi ready"));
  printMessage(MSG_VERSION, F(""));

  // Set Serial Protocol Commands
  setSerialProtocol();

  // Init BL Controller
  initBlController();
  // Init Sinus Arrays
  initMotorStuff();
  
  // switch off PWM Power
  motorPowerOff();
    
  // Read Config, initialize if version does not match or CRC fails
  readEEPROM();
  if (config.versEEPROM != VERSION_EEPROM)
  {
    printMessage(MSG_WARNING, F("EEPROM version mismatch, initialized to default"));
    setDefaultParameters();
    writeEEPROM();
  }
  
  // Start I2C and Configure Frequency
  Wire.begin();
  TWSR = 0;                                  // no prescaler => prescaler = 1
  TWBR = ((16000000L / I2C_SPEED) - 16) / 2; // change the I2C clock rate
  TWCR = 1<<TWEN;                            // enable twi module, no interrupt
 
  // Initialize MPU 
  initResolutionDevider();
    
  // init I2C and MPU6050
  if (initI2C()) {  
    // Init IMU variables
    initIMU();
    // Gyro Offset calibration
    if (config.gyroCal) {
      gyroCalibrateCmd();
    }
  } else {
 //   gimState = GIM_ERROR;
  }
  
  // set sensor orientation
  initSensorOrientation();
  
  // Init PIDs parameters
  initPIDs();

  // init RC variables
  initRC();

  // Init RC-Input
  initRCPins();

  LEDPIN_OFF
  LEDPIN2_OFF

	gyroOffsettemp_PITCH	= config.gyrOffsetY;
	gyroOffsettemp_ROLL 	= config.gyrOffsetX;
//	gyroOffsettemp_YAW		= config.gyrOffsetZ;

  CH2_OFF
  CH3_OFF

}

/************************/
/* PID Controller       */
/************************/
// PID integer inplementation
//   DTms  ... sample period (ms)
//   DTinv ... sample frequency (Hz), inverse of DT (just to avoid division)
int32_t ComputePID(int32_t DTms, int32_t DTinv, int32_t in, int32_t setPoint, int32_t *errorSum, int32_t *errorOld, int32_t Kp, int16_t Ki, int32_t Kd)
{
  int32_t error = setPoint - in;
  int32_t Ierr;
   
  Ierr = error * Ki * DTms;
  Ierr = constrain_int32(Ierr, -(int32_t)1000*100, (int32_t)1000*100);
  *errorSum += Ierr;
 
  /*Compute PID Output*/
  int32_t out = (Kp * error) + *errorSum + Kd * (error - *errorOld) * DTinv;
  *errorOld = error;

  out = out / 4096 / 8;
  
  return out;
  
}


/*****************************************************************

main loop execution time budget


exected each iteration (main tick)
  time(us)   function
  --------------------------------
   330        readGyros
   175        updateGyroAttitude
    21        updateACCAttitude
   372        getAttiduteAngles
    92        pid pitch
    92        pid roll
     6        motor update
    84        RC low pass
  --------------------------------
  1172        sum
  1250        measured, real, no RC

executed each 10th iteration (sub tick)
  time(us)   function
  --------------------------------
   330        readACC
   120        updateACC
   210        voltage compensation
    56        gimbal state
    26        RC roll
    26        RC pitch
   286        evaluate RC
  --------------------------------
   330        maximum of all
   --        measured, real, no RC
  --------------------------------


total
  time(us)    function
  --------------------------------
   1172       main tick execution time
    330       sub tick maximum
  --------------------------------
   1502       sum 
  ================================


motor ISR duration
  version r161
  10 us every 31 us
  18 us every 2000 us
  
  --> improved
   6.3 us every 31 us
   
RC interrupt duration
   43 us, every 1000 ms (one RC channel)
   54 us, every 1000 ms, worst case, two RC channel

*******************************************************************/

void restart() 
{
		gimState = GIM_UNLOCKED;
		handheld_ok = 0;
		startwait = 0;
		stateCount = 0;
		FollowP = 0;
		FollowR = 0;
		FollowP_MoveTotal = 0;
//		FollowR_MoveTotal = 0;
		FollowP_Move = 0;
		FollowR_Move = 0;
//		FollowR_MoveOLD = 0;
		FollowR_Handle = 0;
		FollowR_FAplus = 0;
		LPFaccADC_R = 0;
		pitchAngleSet = 0;
		PitchPhiSet = 0;
		qLPPitch[0] = 0;
		qLPPitch[1] = 0;
		qLPPitch[2] = 0;
}
/**********************************************/
/* Main Loop                                  */
/**********************************************/
void loop() 
{ 
  int32_t pitchPIDVal;
  int32_t rollPIDVal;
  
  static char pOutCnt = 0;
  static char tOutCnt = 0;
  static char tOutCntSub = 0;
  static int stateCount = 0;
  static uint8_t ledBlinkCnt = 0;
  static uint8_t ledBlinkOnTime = 6;
  static uint8_t ledBlinkPeriod = 12;
//  static uint8_t ledBlinkCnt2 = 0;
//  static uint8_t ledBlinkOnTime2 = 5;
//  static uint8_t ledBlinkPeriod2 = 10;

  if (motorUpdate) // loop runs with motor ISR update rate (500 Hz)
  {
   
    motorUpdate = false;

    CH2_ON
    
    // loop period
    //     2.053/2.035 ms max/min, error = +5/-13 us (w/o rc)
    //     2.098/2.003 ms max/min, error = +50/-45 us (1 x PPM16 1 x PWM)
    
    // update IMU data            
    readGyros();   // td = 330us

    if (config.enableGyro) updateGyroAttitude(); // td = 176 us
    if (config.enableACC) updateACCAttitude(); // td = 21 us
    getAttiduteAngles(); // td = 372 us
   
    //****************************
    // pitch PID
    //****************************
    if (fpvModeFreezePitch==false) {
      // td = 92 us
//      pitchPIDVal = ComputePID(DT_INT_MS, DT_INT_INV, angle[PITCH], pitchAngleSet*1000, &pitchErrorSum, &pitchErrorOld, pitchPIDpar.Kp, pitchPIDpar.Ki, pitchPIDpar.Kd);
      pitchPIDVal = ComputePID(DT_INT_MS, DT_INT_INV, angle[PITCH], FollowP_MoveTotal, &pitchErrorSum, &pitchErrorOld, pitchPIDpar.Kp, pitchPIDpar.Ki, pitchPIDpar.Kd);
      // motor control
      pitchMotorDrive = pitchPIDVal * config.dirMotorPitch;
    }
 
    //****************************
    // roll PID
    //****************************
    if (fpvModeFreezeRoll==false) {
      // td = 92 us
//      rollPIDVal = ComputePID(DT_INT_MS, DT_INT_INV, angle[ROLL], rollAngleSet*1000, &rollErrorSum, &rollErrorOld, rollPIDpar.Kp, rollPIDpar.Ki, rollPIDpar.Kd);
      rollPIDVal = ComputePID(DT_INT_MS, DT_INT_INV, angle[ROLL], 0, &rollErrorSum, &rollErrorOld, rollPIDpar.Kp, rollPIDpar.Ki, rollPIDpar.Kd);
      // motor control
      rollMotorDrive = rollPIDVal * config.dirMotorRoll;
    }
	//****************************
	// Follow PITCH & ROLL
	//****************************
	utilLP_float(&FollowP, pitchMotorDrive, 0.05);
	utilLP_float(&FollowR,  rollMotorDrive, 0.05);
	if (handheld_ok == 0) {
		utilLP_int(&LPFaccADC_R, accADC[ROLL], 10);
//		utilLP_int(&LPFaccADC_P, accADC[PITCH],5);
//		if (gimState == GIM_UNLOCKED) {
//			ACFC_RatioR = map(constrain(abs(LPFaccADC_R),800,10000), 800, 10000, 100,1000)*0.01;
//			ACFC_RatioP = map(constrain(abs(LPFaccADC_P),800,10000), 800, 10000, 100,1000)*0.01;
//		}
		if (gimState == GIM_LOCKED) {
			startwait += 1;
			if (startwait ==  100) {
				pitchMotorDriveOffset = FollowP;
				rollMotorDriveOffset = FollowR;
			}
			if (startwait >   110) { pitchMotorDrive += 1500; }
			if (startwait ==  200) { 
				handheld_ok = 1;
				startwait = 0;
				stateCount = 0;
			}
		}
	}

	if (abs(gyroADC[PITCH]) > 30000 || abs(gyroADC[ROLL]) > 30000) { restart(); }

	if (handheld_ok == 1) {
		// **** Follow PITCH ****
		if (FollowP_FreeArea < 250) {
			FollowP_Move = pitchMotorDriveOffset - FollowP;
			if        (FollowP_Move >  FollowP_FreeArea) { FollowP_Move -= FollowP_FreeArea;
			} else if (FollowP_Move < -FollowP_FreeArea) { FollowP_Move += FollowP_FreeArea;
			} else { FollowP_Move  = 0; }

			if (config.rcAbsoluteRoll == 1) {
 	 			FollowP_Move = (FollowP_Move*0.4 * abs(FollowP_Move)*0.4+FollowP_Move) / FollowP_Quick;
				FollowP_MoveTotal = FollowP_MoveTotal + FollowP_Move;
				FollowP_Move = 0;
			} else {
			 	FollowP_Move = FollowP_Move*0.1 * abs(FollowP_Move)*0.1 * FollowP_Quick + FollowP_Move;
			}
		} else { 
			FollowP_Move = 0;
			FollowP_MoveTotal = 0;
		}
		// **** Follow ROLL ****
		if (FollowR_FreeArea < 400) {
			FollowR_Move = rollMotorDriveOffset - FollowR;
			if (config.rcAbsoluteRoll == 1) {
				if (abs(FollowP_MoveTotal) > 35000) {
					if (FollowR_Handle == 0) {
						if (	   FollowR_Move >  450) { FollowR_Handle = 1; FollowR_FAplus = -900;
						} else if (FollowR_Move < -450) { FollowR_Handle = 2; FollowR_FAplus =  900;
						} else {						  FollowR_Handle = 3; FollowR_FAplus =   0; }
					}
				} else { FollowR_Handle = 0; FollowR_FAplus = 0; }
				FollowR_Move += FollowR_FAplus;
				FollowR_Move *= (map(constrain(abs(FollowP_MoveTotal), 35000, 45000), 35000, 45000, 0, 100)*0.01);
			}

			if        (FollowR_Move >  FollowR_FreeArea) { FollowR_Move -= FollowR_FreeArea;
			} else if (FollowR_Move < -FollowR_FreeArea) { FollowR_Move += FollowR_FreeArea;
			} else { FollowR_Move  = 0; }

			FollowR_Move = constrain(FollowR_Move, -300, 300);
			FollowR_Move = FollowR_Move*0.1 * abs(FollowR_Move)*0.1 * FollowR_Quick + FollowR_Move;
		} else {
			FollowR_Move = 0;
		}
		// **********************

	} else if (handheld_ok == 2) {
		// **** Follow OFF / Servo Control ****
		if (startwait < 2000) {
			startwait ++;
			PitchPhiSet = PitchPhiSet / 2000 * startwait;
		}
		
		if (fpvModePitch==true) {
    		pitchAngleSet = utilLP3_float(qLPPitch, PitchPhiSet, rcLPFPitchFpv_tc);
		} else if(config.rcAbsolutePitch==1) {
			pitchAngleSet = utilLP3_float(qLPPitch, PitchPhiSet, rcLPFPitch_tc); // 63us
		} else {
			utilLP_float(&pitchAngleSet, PitchPhiSet, 0.01);
		}
		FollowP_MoveTotal = pitchAngleSet*1000;
		FollowP_Move = 0;
		FollowR_Move = 0;
	}

		// **********************
	
	if (handheld_ok == 1) {
		Follow_gyrOffsetX = gyroOffsettemp_ROLL + FollowR_Move;
		Follow_gyrOffsetY = gyroOffsettemp_PITCH + FollowP_Move;
	} else {
		Follow_gyrOffsetX = gyroOffsettemp_ROLL;
		Follow_gyrOffsetY = gyroOffsettemp_PITCH;
	}

	//****************************
	//****************************
	// motor update t=6us (*)
	if (enableMotorUpdates) {
		MoveMotorPosSpeed(config.motorNumberPitch, pitchMotorDrive, maxPWMmotorPitchScaled); // set pitch motor pwm
		MoveMotorPosSpeed(config.motorNumberRoll, rollMotorDrive, maxPWMmotorRollScaled); // set roll motor pwm
	}

	//****************************
	// slow rate actions
	//****************************
	switch (count) {
	case 1: readACCs(); break ;// td = 330us
	case 2: updateACC(); break ;// td = 120us
	case 3: voltageCompensation(); break;// td = 210us, total
	case 4: // gimbal state transitions, td=56us
		switch (gimState) {
			case GIM_IDLE: // wait 2 sec to settle ACC, before PID controlerbecomes active 
				stateCount++;
				if (stateCount >= LOOPUPDATE_FREQ/10*1) { // 1 sec 
					gimState = GIM_UNLOCKED;
					stateCount = 0;
				}
				// allow settling IMU
				enableMotorUpdates = false;
				setACCtc(0.2);
				disableAccGtest = true;
			break;
		
			case GIM_UNLOCKED: // allow PID controller to settle on ACC position
				if (abs(accADC[ROLL]) < 650 && abs(accADC[PITCH]) < 650) {
					stateCount++;
				} else {
					startwait++;
				}
				if (stateCount >= LOOPUPDATE_FREQ/10*1 || startwait >= LOOPUPDATE_FREQ/10*LOCK_TIME_SEC) {
					gimState = GIM_LOCKED;
					stateCount = 0;
					startwait = 0;
				}
				// fast settling of desired position
				enableMotorUpdates = true;
				if (stateCount <= LOOPUPDATE_FREQ/20*1) {
					setACCtc(2.0);
				} else {
					setACCtc(6.0);
				}
				disableAccGtest = true;
				break;
		
			case GIM_LOCKED: // normal operation
				enableMotorUpdates = true;
				disableAccGtest = false;
				if (altModeAccTime) { // alternate time constant mode switch
					setACCtc(config.accTimeConstant2);
				} else {
					setACCtc(config.accTimeConstant);
				}
				break;
		
			case GIM_ERROR: // error state
				enableMotorUpdates = false; // switch off motors
				motorPowerOff(); // switch off PWM Power
				break;
		}
		// handle mode switches
		decodeModeSwitches();  // td = 4 us
		break;

	case 5: // RC Pitch function
		evaluateRCPitch();
		// td = 26 us
		getSetpoint(&PitchPhiSet, RC_DATA_PITCH, RC_DATA_FPV_PITCH, fpvModePitch, config.rcAbsolutePitch, config.maxRCPitch, config.minRCPitch);
		if (abs(PitchPhiSet) > 20 && handheld_ok == 1) { handheld_ok = 2; }
		break;

	case 6: // RC roll function
//		evaluateRCRoll();
		// td = 26us
//		getSetpoint(&RollPhiSet, RC_DATA_ROLL, RC_DATA_FPV_ROLL, fpvModeRoll, config.rcAbsoluteRoll, config.maxRCRoll, config.minRCRoll);
		break;

	case 7: // evaluate RC-Signals. td = 90 us
//		evaluateRCAux();
		// check RC channel timeouts
		checkRcTimeouts();  // td = 15 us
		break;

	case 8: // read RC Anlog inputs
//		readRCAnalog(); // td = 354 us (if all 3 enabled, 118 us each analog channel)
		FollowP_FreeArea	= abs(config.minRCRoll)*5;
		if (config.rcAbsoluteRoll == 1) {
			FollowP_Quick		= abs(config.maxRCRoll)+1;
		} else {
			FollowP_Quick		= abs(config.maxRCRoll)*3+1;
		}
		FollowR_FreeArea	= abs(config.rcGainRoll)*2;
		FollowR_Quick		= config.rcLPFRoll/3+1;
		break;

	case 9: //   regular i2c test
		mpu.testConnection();
		break;

	case 10: // regular ACC output
		pOutCnt++;
		if (pOutCnt == (LOOPUPDATE_FREQ/10/POUT_FREQ)) {
			if (config.fTrace != TRC_OFF) { printTrace(config.fTrace); }
			pOutCnt = 0;
		}
      // print regular trace output
		tOutCnt++;
		if (tOutCnt == (LOOPUPDATE_FREQ/10/TRACE_OUT_FREQ)) {
			tOutCntSub++;
			if (tOutCntSub >= TRC_LAST_IDX) { tOutCntSub = 1; }
			if (config.sTrace == TRC_ALL) {
				// cycle all trace types
				printTrace((traceModeType)tOutCntSub);
			} else if (config.sTrace != TRC_OFF) {
				// use specific trace type
				printTrace(config.sTrace);
			}
			tOutCnt = 0;
		}

		if (handheld_ok == 0) {
			ledBlinkCnt += 1;
			if (ledBlinkCnt <= ledBlinkOnTime) {
				LEDPIN_ON
				LEDPIN2_ON
			} else if (ledBlinkCnt <= ledBlinkPeriod) {
				LEDPIN_OFF
				LEDPIN2_OFF
			} else {
				ledBlinkCnt = 0;
			}
//		} else if (FollowR_Move != 0) {
		} else if (abs(accADC[ROLL]) > 700 || abs(accADC[PITCH]) > 700) {
			ledBlinkCnt += 3;
			if (ledBlinkCnt <= ledBlinkOnTime) {
				LEDPIN_ON
				LEDPIN2_ON
			} else if (ledBlinkCnt <= ledBlinkPeriod) {
				LEDPIN_OFF
				LEDPIN2_OFF
			} else {
				ledBlinkCnt = 0;
			}
		} else {
			LEDPIN_ON
			LEDPIN2_ON
			ledBlinkCnt = 0;
		}

#ifdef STACKHEAPCHECK_ENABLE
		stackHeapEval(false);
#endif
		count=0;
		break;
	default:
		break;
	}
	count++;

	//****************************
	// Evaluate Serial inputs 
	//****************************
	sCmd.readSerial();


	// worst-case finalize after
	//    1.67 ms (w/o RC)
	//    1.76 ms (with 1 x PPM)
	//    1.9 ms (with 2 RC channels + 1 x PPM )

	CH2_OFF
}

}


